from typing import Dict, List

from sqlalchemy.engine import Engine, LegacyCursorResult, create_engine  # type: ignore

from fides.api.graph.execution import ExecutionNode
from fides.api.schemas.connection_configuration.connection_secrets_mysql import (
    MySQLSchema,
    MySQLSSLMode,
)
from fides.api.service.connectors.query_configs.mysql_query_config import (
    MySQLQueryConfig,
)
from fides.api.service.connectors.query_configs.query_config import SQLQueryConfig
from fides.api.service.connectors.sql_connector import SQLConnector
from fides.api.util.collection_util import Row
from fides.config import get_config

CONFIG = get_config()


class MySQLConnector(SQLConnector):
    """Connector specific to MySQL"""

    secrets_schema = MySQLSchema

    def build_uri(self) -> str:
        """Build URI of format mysql+pymysql://[user[:password]@][netloc][:port][/dbname]"""
        config = self.secrets_schema(**self.configuration.secrets or {})

        user_password = ""
        if config.username:
            user = config.username
            password = f":{config.password}" if config.password else ""
            user_password = f"{user}{password}@"

        netloc = config.host
        port = f":{config.port}" if config.port else ""
        dbname = f"/{config.dbname}" if config.dbname else ""
        url = f"mysql+pymysql://{user_password}{netloc}{port}{dbname}"
        return url

    def build_ssh_uri(self, local_address: tuple) -> str:
        """Build URI of format mysql+pymysql://[user[:password]@][ssh_host][:ssh_port][/dbname]"""
        config = self.secrets_schema(**self.configuration.secrets or {})

        user_password = ""
        if config.username:
            user = config.username
            password = f":{config.password}" if config.password else ""
            user_password = f"{user}{password}@"

        local_host, local_port = local_address
        netloc = local_host
        port = f":{local_port}" if local_port else ""
        dbname = f"/{config.dbname}" if config.dbname else ""
        url = f"mysql+pymysql://{user_password}{netloc}{port}{dbname}"
        return url

    # Overrides SQLConnector.create_client
    def create_client(self) -> Engine:
        """Returns a SQLAlchemy Engine that can be used to interact with a database"""
        if (
            self.configuration.secrets
            and self.configuration.secrets.get("ssh_required", False)
            and CONFIG.security.bastion_server_ssh_private_key
        ):
            config = self.secrets_schema(**self.configuration.secrets or {})
            self.create_ssh_tunnel(host=config.host, port=config.port)
            self.ssh_server.start()
            uri = self.build_ssh_uri(local_address=self.ssh_server.local_bind_address)
        else:
            uri = (self.configuration.secrets or {}).get("url") or self.build_uri()
        connect_args = self.get_connect_args()
        return create_engine(
            uri,
            hide_parameters=self.hide_parameters,
            echo=not self.hide_parameters,
            connect_args=connect_args,
        )

    def query_config(self, node: ExecutionNode) -> SQLQueryConfig:
        """Query wrapper corresponding to the input execution_node."""
        return MySQLQueryConfig(node)

    def get_connect_args(self) -> Dict[str, Dict[str, MySQLSSLMode]]:
        """Get connection arguments for the engine"""

        connect_args = {}

        ssl_mode = self.configuration.secrets.get("ssl_mode", MySQLSSLMode.preferred)

        # if SSL mode was assigned in the connector configuration
        # otherwise connect_args remains empty
        if ssl_mode:
            connect_args["ssl"] = {"mode": ssl_mode}

        return connect_args

    @staticmethod
    def cursor_result_to_rows(results: LegacyCursorResult) -> List[Row]:
        """
        Convert SQLAlchemy results to a list of dictionaries
        """
        return SQLConnector.default_cursor_result_to_rows(results)



"""add consent automation tables

Revision ID: d69cf8f82a58
Revises: fc82ab64bd5e
Create Date: 2024-07-24 23:09:46.681097

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "d69cf8f82a58"
down_revision = "fc82ab64bd5e"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "plus_consent_automation",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("connection_config_id", sa.String(), nullable=False),
        sa.ForeignKeyConstraint(
            ["connection_config_id"], ["connectionconfig.id"], ondelete="CASCADE"
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("connection_config_id"),
    )
    op.create_index(
        op.f("ix_plus_consent_automation_id"),
        "plus_consent_automation",
        ["id"],
        unique=False,
    )
    op.create_table(
        "plus_consentable_item",
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column("consent_automation_id", sa.String(), nullable=False),
        sa.Column("external_id", sa.String(), nullable=False),
        sa.Column("parent_id", sa.String(), nullable=True),
        sa.Column("notice_id", sa.String(), nullable=True),
        sa.Column("type", sa.String(), nullable=False),
        sa.Column("name", sa.String(), nullable=False),
        sa.ForeignKeyConstraint(
            ["consent_automation_id"],
            ["plus_consent_automation.id"],
            ondelete="CASCADE",
        ),
        sa.ForeignKeyConstraint(
            ["notice_id"],
            ["privacynotice.id"],
        ),
        sa.ForeignKeyConstraint(
            ["parent_id"], ["plus_consentable_item.id"], ondelete="CASCADE"
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("consent_automation_id", "type", "external_id"),
    )
    op.create_index(
        op.f("ix_plus_consentable_item_id"),
        "plus_consentable_item",
        ["id"],
        unique=False,
    )


def downgrade():
    op.drop_index(
        op.f("ix_plus_consentable_item_id"), table_name="plus_consentable_item"
    )
    op.drop_table("plus_consentable_item")
    op.drop_index(
        op.f("ix_plus_consent_automation_id"), table_name="plus_consent_automation"
    )
    op.drop_table("plus_consent_automation")



"""add detection discovery tables

Revision ID: fc2b2c06e595
Revises: d873e3e430b0
Create Date: 2024-04-25 19:53:59.562332

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "fc2b2c06e595"
down_revision = "d873e3e430b0"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    # Staged resource table
    op.create_table(
        "stagedresource",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("name", sa.String(), nullable=True),
        sa.Column("urn", sa.String(), nullable=False),
        sa.Column("resource_type", sa.String(), nullable=True),
        sa.Column("description", sa.String(), nullable=True),
        sa.Column("monitor_config_id", sa.String(), nullable=True),
        sa.Column("source_modified", sa.DateTime(timezone=True), nullable=True),
        sa.Column(
            "classifications",
            sa.ARRAY(postgresql.JSONB(astext_type=sa.Text())),
            server_default="{}",
            nullable=False,
        ),
        sa.Column(
            "user_assigned_data_categories",
            sa.ARRAY(sa.String()),
            server_default="{}",
            nullable=False,
        ),
        sa.Column(
            "children", sa.ARRAY(sa.String()), server_default="{}", nullable=False
        ),
        sa.Column("parent", sa.String(), nullable=True),
        sa.Column("diff_status", sa.String(), nullable=True),
        sa.Column(
            "child_diff_statuses",
            postgresql.JSONB(astext_type=sa.Text()),
            server_default="{}",
            nullable=False,
        ),
        sa.Column(
            "meta",
            postgresql.JSONB(astext_type=sa.Text()),
            server_default="{}",
            nullable=False,
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_stagedresource_id"), "stagedresource", ["id"], unique=False
    )
    op.create_index(
        op.f("ix_stagedresource_urn"), "stagedresource", ["urn"], unique=True
    )
    op.create_index(
        op.f("ix_stagedresource_resource_type"),
        "stagedresource",
        ["resource_type"],
        unique=False,
    )

    # Monitor config table
    op.create_table(
        "monitorconfig",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("name", sa.String(), nullable=True),
        sa.Column("key", sa.String(), nullable=False),
        sa.Column("connection_config_id", sa.String(), nullable=False),
        sa.Column(
            "classify_params", postgresql.JSONB(astext_type=sa.Text()), nullable=True
        ),
        sa.ForeignKeyConstraint(
            ["connection_config_id"],
            ["connectionconfig.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(op.f("ix_monitorconfig_id"), "monitorconfig", ["id"], unique=False)
    op.create_index(op.f("ix_monitorconfig_key"), "monitorconfig", ["key"], unique=True)
    op.create_index(
        op.f("ix_monitorconfig_connection_config_id"),
        "monitorconfig",
        ["connection_config_id"],
        unique=False,
    )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    # Staged resource table
    op.drop_index(op.f("ix_stagedresource_resource_type"), table_name="stagedresource")
    op.drop_index(op.f("ix_stagedresource_urn"), table_name="stagedresource")
    op.drop_index(op.f("ix_stagedresource_id"), table_name="stagedresource")
    op.drop_table("stagedresource")

    # Monitor config table
    op.drop_index(op.f("ix_monitorconfig_id"), table_name="monitorconfig")
    op.drop_index(op.f("ix_monitorconfig_key"), table_name="monitorconfig")
    op.drop_index(
        op.f("ix_monitorconfig_connection_config_id"), table_name="monitorconfig"
    )
    op.drop_table("monitorconfig")
    # ### end Alembic commands ###



"""add property config and css

Revision ID: 0debabbb9c6a
Revises: 4fba24045fec
Create Date: 2024-05-06 16:45:45.958888

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "0debabbb9c6a"
down_revision = "4fba24045fec"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "plus_property",
        sa.Column(
            "privacy_center_config",
            postgresql.JSONB(astext_type=sa.Text()),
            nullable=True,
        ),
    )
    op.add_column("plus_property", sa.Column("stylesheet", sa.Text(), nullable=True))
    op.add_column(
        "plus_property",
        sa.Column("paths", sa.ARRAY(sa.String()), server_default="{}", nullable=False),
    )


def downgrade():
    op.drop_column("plus_property", "paths")
    op.drop_column("plus_property", "stylesheet")
    op.drop_column("plus_property", "privacy_center_config")



from sqlalchemy import Column, ForeignKey, String

from fides.api.db.base_class import Base


class SystemManager(Base):
    """Many to many table to link users to systems as system managers.

    Usually access via user.systems or system.users instead of via this table directly.
    """

    user_id = Column(
        String, ForeignKey("fidesuser.id", ondelete="CASCADE"), primary_key=True
    )
    system_id = Column(
        String, ForeignKey("ctl_systems.id", ondelete="CASCADE"), primary_key=True
    )



"""add_callback_succeeded

Revision ID: 4b2eade4353c
Revises: d7feff990012
Create Date: 2024-05-02 19:38:46.537231

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "4b2eade4353c"
down_revision = "d7feff990012"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "requesttask", sa.Column("callback_succeeded", sa.Boolean(), nullable=True)
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("requesttask", "callback_succeeded")



from typing import List, Optional

from fideslang.validation import FidesKey
from pydantic import ConfigDict

from fides.api.models.policy import WebhookDirection
from fides.api.schemas.base_class import FidesSchema
from fides.api.schemas.connection_configuration.connection_config import (
    ConnectionConfigurationResponse,
)


class WebhookBase(FidesSchema):
    """Base schema for Webhooks"""

    direction: WebhookDirection
    key: Optional[FidesKey] = None
    name: Optional[str] = None


class PolicyWebhookCreate(WebhookBase):
    """Request schema for creating/updating a Policy Webhook"""

    connection_config_key: FidesKey
    model_config = ConfigDict(use_enum_values=True)


class PolicyWebhookResponse(WebhookBase):
    """Response schema after creating a PolicyWebhook"""

    connection_config: Optional[ConnectionConfigurationResponse] = None
    order: int
    model_config = ConfigDict(from_attributes=True)


class PolicyWebhookUpdate(FidesSchema):
    """Request schema for updating a single webhook - fields are optional"""

    direction: Optional[WebhookDirection] = None
    name: Optional[str] = None
    connection_config_key: Optional[FidesKey] = None
    order: Optional[int] = None
    model_config = ConfigDict(
        from_attributes=True, extra="forbid", use_enum_values=True
    )


class WebhookOrder(FidesSchema):
    """Schema for displaying a minimal amount of information about the webhook and its order"""

    key: FidesKey
    order: int
    model_config = ConfigDict(from_attributes=True)


class PolicyWebhookUpdateResponse(FidesSchema):
    """Response schema after a PATCH to a single webhook - because updating the order of this webhook can update the
    order of other webhooks, new_order will include the new order if order was adjusted at all
    """

    resource: PolicyWebhookResponse
    new_order: List[WebhookOrder]


class PolicyWebhookDeleteResponse(FidesSchema):
    """Response schema after deleting a webhook; new_order includes remaining reordered webhooks if applicable"""

    new_order: List[WebhookOrder]
    model_config = ConfigDict(from_attributes=True)



from typing import Annotated, List

from fastapi import Body, Depends, Security
from fastapi_pagination import Page, Params
from fastapi_pagination.bases import AbstractPage
from fastapi_pagination.ext.sqlalchemy import paginate
from fideslang.validation import FidesKey
from loguru import logger
from pydantic import Field
from sqlalchemy.orm import Session
from starlette.exceptions import HTTPException
from starlette.status import HTTP_200_OK, HTTP_400_BAD_REQUEST, HTTP_404_NOT_FOUND

from fides.api.api import deps
from fides.api.api.v1.endpoints.policy_endpoints import get_policy_or_error
from fides.api.common_exceptions import KeyOrNameAlreadyExists, WebhookOrderException
from fides.api.db.base_class import get_key_from_data
from fides.api.models.policy import (
    Policy,
    PolicyPostWebhook,
    PolicyPreWebhook,
    WebhookTypes,
)
from fides.api.oauth.utils import verify_oauth_client
from fides.api.schemas import policy_webhooks as schemas
from fides.api.schemas.policy_webhooks import PolicyWebhookDeleteResponse
from fides.api.util.api_router import APIRouter
from fides.api.util.connection_util import get_connection_config_or_error
from fides.common.api import scope_registry as scopes
from fides.common.api.v1 import urn_registry as urls

router = APIRouter(tags=["DSR Policy Webhooks"], prefix=urls.V1_URL_PREFIX)


@router.get(
    urls.POLICY_WEBHOOKS_PRE,
    status_code=HTTP_200_OK,
    response_model=Page[schemas.PolicyWebhookResponse],
    dependencies=[Security(verify_oauth_client, scopes=[scopes.WEBHOOK_READ])],
)
def get_policy_pre_execution_webhooks(
    *,
    db: Session = Depends(deps.get_db),
    policy_key: FidesKey,
    params: Params = Depends(),
) -> AbstractPage[PolicyPreWebhook]:
    """
    Return a paginated list of all Pre-Execution Webhooks that will
    run in order for the Policy **before** a Privacy Request is executed.
    """
    policy = get_policy_or_error(db, policy_key)

    logger.debug(
        "Finding all Pre-Execution Webhooks for Policy '{}' with pagination params '{}'",
        policy.key,
        params,
    )
    return paginate(policy.pre_execution_webhooks.order_by("order"), params=params)  # type: ignore[attr-defined]


@router.get(
    urls.POLICY_WEBHOOKS_POST,
    status_code=HTTP_200_OK,
    response_model=Page[schemas.PolicyWebhookResponse],
    dependencies=[Security(verify_oauth_client, scopes=[scopes.WEBHOOK_READ])],
)
def get_policy_post_execution_webhooks(
    *,
    db: Session = Depends(deps.get_db),
    policy_key: FidesKey,
    params: Params = Depends(),
) -> AbstractPage[PolicyPostWebhook]:
    """
    Return a paginated list of all Post-Execution Webhooks that will run in order for the Policy **after** a
    Privacy Request is executed.
    """
    policy = get_policy_or_error(db, policy_key)

    logger.debug(
        "Finding all Post-Execution Webhooks for Policy '{}' with pagination params '{}'",
        policy.key,
        params,
    )
    return paginate(policy.post_execution_webhooks.order_by("order"), params=params)  # type: ignore[attr-defined]


def put_webhooks(
    webhook_cls: WebhookTypes,
    policy_key: FidesKey,
    db: Session = Depends(deps.get_db),
    webhooks: List[schemas.PolicyWebhookCreate] = Body(...),
) -> List[WebhookTypes]:
    """
    Helper method to PUT pre-execution or post-execution policy webhooks.

    Creates/updates webhooks with the same "order" in which they arrived. This endpoint is all-or-nothing.
    Either all webhooks should be created/updated or none should be updated. Deletes any webhooks not present
    in the request body.
    """
    policy = get_policy_or_error(db, policy_key)

    keys = [
        get_key_from_data(webhook.model_dump(mode="json"), type(webhook_cls).__name__)
        for webhook in webhooks
    ]
    names = [webhook.name for webhook in webhooks]
    # Because resources are dependent on each other for order, we want to make sure that we don't have multiple
    # resources in the request that actually point to the same object.
    if len(keys) != len(set(keys)) or len(names) != len(set(names)):
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail="Check request body: there are multiple webhooks whose keys or names resolve to the same value.",
        )

    staged_webhooks = []  # Webhooks will be committed at the end
    for webhook_index, schema in enumerate(webhooks):
        connection_config = get_connection_config_or_error(
            db, schema.connection_config_key
        )

        try:
            webhook = webhook_cls.create_or_update(
                db=db,
                data={
                    "key": schema.key,
                    "name": schema.name,
                    "policy_id": policy.id,
                    "connection_config_id": connection_config.id,
                    "direction": schema.direction,
                    "order": webhook_index,  # Add in the order they arrived in the request
                },
            )
            staged_webhooks.append(webhook)
        except KeyOrNameAlreadyExists as exc:
            raise HTTPException(
                status_code=HTTP_400_BAD_REQUEST,
                detail=exc.args[0],
            )

    staged_webhook_keys = [webhook.key for webhook in staged_webhooks]
    webhooks_to_remove = getattr(
        policy, f"{webhook_cls.prefix}_execution_webhooks"
    ).filter(
        webhook_cls.key.not_in(staged_webhook_keys)  # type: ignore
    )

    if webhooks_to_remove.count():
        logger.info(
            "Removing {}-Execution Webhooks from Policy '{}' that were not included in request: {}",
            webhook_cls.prefix.capitalize(),
            policy.key,
            [webhook.key for webhook in webhooks_to_remove],
        )
        webhooks_to_remove.delete()

    logger.info(
        "Creating/updating Policy Pre-Execution Webhooks: {}", staged_webhook_keys
    )
    # Committing to database now, as a last step, once we've verified that all the webhooks
    # in the request are free of issues.
    db.commit()
    return staged_webhooks


@router.put(
    urls.POLICY_WEBHOOKS_PRE,
    status_code=HTTP_200_OK,
    dependencies=[
        Security(verify_oauth_client, scopes=[scopes.WEBHOOK_CREATE_OR_UPDATE])
    ],
    response_model=List[schemas.PolicyWebhookResponse],
)
def create_or_update_pre_execution_webhooks(
    *,
    policy_key: FidesKey,
    db: Session = Depends(deps.get_db),
    webhooks: Annotated[List[schemas.PolicyWebhookCreate], Field(max_length=50)],  # type: ignore
) -> List[PolicyPreWebhook]:
    """
    Create or update the list of Policy Pre-Execution Webhooks that run **before** query execution.

    All webhooks must be included in the request in the desired order. Any missing webhooks
    from the request body will be removed.
    """
    return put_webhooks(PolicyPreWebhook, policy_key, db, webhooks)  # type: ignore


@router.put(
    urls.POLICY_WEBHOOKS_POST,
    status_code=HTTP_200_OK,
    dependencies=[
        Security(verify_oauth_client, scopes=[scopes.WEBHOOK_CREATE_OR_UPDATE])
    ],
    response_model=List[schemas.PolicyWebhookResponse],
)
def create_or_update_post_execution_webhooks(
    *,
    policy_key: FidesKey,
    db: Session = Depends(deps.get_db),
    webhooks: Annotated[List[schemas.PolicyWebhookCreate], Field(max_length=50)],  # type: ignore
) -> List[PolicyPostWebhook]:
    """
    Create or update the list of Policy Post-Execution Webhooks that run **after** query execution.

    All webhooks must be included in the request in the desired order. Any missing webhooks
    from the request body will be removed.
    """
    return put_webhooks(PolicyPostWebhook, policy_key, db, webhooks)  # type: ignore


def get_policy_webhook_or_error(
    db: Session,
    policy: Policy,
    webhook_key: FidesKey,
    webhook_cls: WebhookTypes,
) -> WebhookTypes:
    """Helper method to load a Pre-Execution or Post-Execution Policy Webhook or 404

    Also verifies that the webhook belongs to the given Policy.
    """
    logger.debug(
        "Finding {}-Execution Webhook with key '{}' for Policy '{}'",
        webhook_cls.prefix.capitalize(),
        webhook_key,
        policy.key,
    )
    loaded_webhook = webhook_cls.filter(
        db=db,
        conditions=(
            (webhook_cls.policy_id == policy.id) & (webhook_cls.key == webhook_key)
        ),
    ).first()
    if not loaded_webhook:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No {webhook_cls.prefix.capitalize()}-Execution Webhook found for key '{webhook_key}' on Policy '{policy.key}'.",
        )
    return loaded_webhook


@router.get(
    urls.POLICY_PRE_WEBHOOK_DETAIL,
    status_code=HTTP_200_OK,
    response_model=schemas.PolicyWebhookResponse,
    dependencies=[Security(verify_oauth_client, scopes=[scopes.WEBHOOK_READ])],
)
def get_policy_pre_execution_webhook(
    *,
    db: Session = Depends(deps.get_db),
    policy_key: FidesKey,
    pre_webhook_key: FidesKey,
) -> PolicyPreWebhook:
    """
    Loads the given Pre-Execution Webhook on the Policy
    """
    policy = get_policy_or_error(db, policy_key)
    return get_policy_webhook_or_error(db, policy, pre_webhook_key, PolicyPreWebhook)  # type: ignore


@router.get(
    urls.POLICY_POST_WEBHOOK_DETAIL,
    status_code=HTTP_200_OK,
    response_model=schemas.PolicyWebhookResponse,
    dependencies=[Security(verify_oauth_client, scopes=[scopes.WEBHOOK_READ])],
)
def get_policy_post_execution_webhook(
    *,
    db: Session = Depends(deps.get_db),
    policy_key: FidesKey,
    post_webhook_key: FidesKey,
) -> PolicyPostWebhook:
    """
    Loads the given Post-Execution Webhook on the Policy
    """
    policy = get_policy_or_error(db, policy_key)
    return get_policy_webhook_or_error(db, policy, post_webhook_key, PolicyPostWebhook)  # type: ignore


def _patch_webhook(
    *,
    db: Session = Depends(deps.get_db),
    policy_key: FidesKey,
    webhook_key: FidesKey,
    webhook_body: schemas.PolicyWebhookUpdate = Body(...),
    webhook_cls: WebhookTypes,
) -> schemas.PolicyWebhookUpdateResponse:
    """Helper method for PATCHing a single webhook, either Pre-Execution or Post-Execution

    If the order is updated, this will affect the order of other webhooks.
    """
    policy = get_policy_or_error(db, policy_key)
    loaded_webhook = get_policy_webhook_or_error(db, policy, webhook_key, webhook_cls)
    data = webhook_body.model_dump(exclude_none=True)

    if data.get("connection_config_key"):
        connection_config = get_connection_config_or_error(
            db, data.get("connection_config_key")  # type: ignore
        )
        data["connection_config_id"] = connection_config.id

    # Removing index from incoming data - we'll set this at the end.
    index = data.pop("order", None)

    try:
        logger.info(
            "Updating {}-Execution Webhook with key '{}' on Policy '{}' ",
            webhook_cls.prefix.capitalize(),
            webhook_key,
            policy_key,
        )
        loaded_webhook.update(db, data=data)
    except KeyOrNameAlreadyExists as exc:
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail=exc.args[0],
        )

    if index is not None and index != loaded_webhook.order:
        logger.info(
            "Reordering {}-Execution Webhooks for Policy '{}'",
            webhook_cls.prefix.capitalize(),
            policy_key,
        )
        try:
            loaded_webhook.reorder_related_webhooks(db=db, new_index=index)
        except WebhookOrderException as exc:
            raise HTTPException(
                status_code=HTTP_400_BAD_REQUEST,
                detail=exc.args[0],
            )

        return schemas.PolicyWebhookUpdateResponse(
            resource=loaded_webhook,
            new_order=[
                webhook
                for webhook in getattr(
                    policy, f"{webhook_cls.prefix}_execution_webhooks"
                ).order_by(webhook_cls.order)
            ],
        )

    # Policy Webhooks are not committed by default, so we commit at the end.
    db.commit()
    return schemas.PolicyWebhookUpdateResponse(resource=loaded_webhook, new_order=[])


@router.patch(
    urls.POLICY_PRE_WEBHOOK_DETAIL,
    status_code=HTTP_200_OK,
    dependencies=[
        Security(verify_oauth_client, scopes=[scopes.WEBHOOK_CREATE_OR_UPDATE])
    ],
    response_model=schemas.PolicyWebhookUpdateResponse,
)
def update_pre_execution_webhook(
    *,
    db: Session = Depends(deps.get_db),
    policy_key: FidesKey,
    pre_webhook_key: FidesKey,
    webhook_body: schemas.PolicyWebhookUpdate = Body(...),
) -> schemas.PolicyWebhookUpdateResponse:
    """PATCH a single Policy Pre-Execution Webhook that runs **prior** to executing the Privacy Request.

    Note that updates to the webhook's "order" can affect the order of the other pre-execution webhooks.
    """
    return _patch_webhook(
        db=db,
        policy_key=policy_key,
        webhook_key=pre_webhook_key,
        webhook_body=webhook_body,
        webhook_cls=PolicyPreWebhook,  # type: ignore
    )


@router.patch(
    urls.POLICY_POST_WEBHOOK_DETAIL,
    status_code=HTTP_200_OK,
    dependencies=[
        Security(verify_oauth_client, scopes=[scopes.WEBHOOK_CREATE_OR_UPDATE])
    ],
    response_model=schemas.PolicyWebhookUpdateResponse,
)
def update_post_execution_webhook(
    *,
    db: Session = Depends(deps.get_db),
    policy_key: FidesKey,
    post_webhook_key: FidesKey,
    webhook_body: schemas.PolicyWebhookUpdate = Body(...),
) -> schemas.PolicyWebhookUpdateResponse:
    """PATCH a single Policy Post-Execution Webhook that runs **after** executing the Privacy Request.

    Note that updates to the webhook's "order" can affect the order of the other post-execution webhooks.
    """
    return _patch_webhook(
        db=db,
        policy_key=policy_key,
        webhook_key=post_webhook_key,
        webhook_body=webhook_body,
        webhook_cls=PolicyPostWebhook,  # type: ignore
    )


def delete_webhook(
    *,
    db: Session = Depends(deps.get_db),
    policy_key: FidesKey,
    webhook_key: FidesKey,
    webhook_cls: WebhookTypes,
) -> PolicyWebhookDeleteResponse:
    """Handles deleting Pre- or Post-Execution Policy Webhooks. Related webhooks are reordered as necessary"""
    policy = get_policy_or_error(db, policy_key)
    loaded_webhook = get_policy_webhook_or_error(db, policy, webhook_key, webhook_cls)
    total_webhook_count = (
        getattr(policy, f"{webhook_cls.prefix}_execution_webhooks").count() - 1
    )
    reordering = total_webhook_count != loaded_webhook.order

    if reordering:
        # Move the webhook to the end and shuffle other webhooks
        logger.info(
            "Reordering {}-Execution Webhooks for Policy '{}'",
            webhook_cls.prefix.capitalize(),
            policy_key,
        )
        loaded_webhook.reorder_related_webhooks(db=db, new_index=total_webhook_count)

    logger.info(
        "Deleting {}-Execution Webhook with key '{}' off of Policy '{}'",
        webhook_cls.prefix.capitalize(),
        webhook_key,
        policy_key,
    )
    loaded_webhook.delete(db=db)
    return PolicyWebhookDeleteResponse(
        new_order=(
            [
                webhook
                for webhook in getattr(
                    policy, f"{webhook_cls.prefix}_execution_webhooks"
                ).order_by(webhook_cls.order)
            ]
            if reordering
            else []
        )
    )


@router.delete(
    urls.POLICY_PRE_WEBHOOK_DETAIL,
    status_code=HTTP_200_OK,
    dependencies=[Security(verify_oauth_client, scopes=[scopes.WEBHOOK_DELETE])],
    response_model=schemas.PolicyWebhookDeleteResponse,
)
def delete_pre_execution_webhook(
    *,
    db: Session = Depends(deps.get_db),
    policy_key: FidesKey,
    pre_webhook_key: FidesKey,
) -> schemas.PolicyWebhookDeleteResponse:
    """Delete the Pre-Execution Webhook from the Policy and reorder remaining webhooks as necessary."""
    return delete_webhook(
        db=db,
        policy_key=policy_key,
        webhook_key=pre_webhook_key,
        webhook_cls=PolicyPreWebhook,  # type: ignore
    )


@router.delete(
    urls.POLICY_POST_WEBHOOK_DETAIL,
    status_code=HTTP_200_OK,
    dependencies=[Security(verify_oauth_client, scopes=[scopes.WEBHOOK_DELETE])],
    response_model=schemas.PolicyWebhookDeleteResponse,
)
def delete_post_execution_webhook(
    *,
    db: Session = Depends(deps.get_db),
    policy_key: FidesKey,
    post_webhook_key: FidesKey,
) -> schemas.PolicyWebhookDeleteResponse:
    """Delete the Post-Execution Webhook from the Policy and reorder remaining webhooks as necessary."""
    return delete_webhook(
        db=db,
        policy_key=policy_key,
        webhook_key=post_webhook_key,
        webhook_cls=PolicyPostWebhook,  # type: ignore
    )



from typing import Literal, Set, Tuple

from fides.api.schemas.connection_configuration import secrets_schemas
from fides.api.schemas.namespace_meta.namespace_meta import NamespaceMeta


class BigQueryNamespaceMeta(NamespaceMeta):
    """
    Represents the namespace structure for BigQuery queries.

    Attributes:
        project_id (str): The ID of the Google Cloud project.
        dataset_id (str): The ID of the BigQuery dataset.
    """

    connection_type: Literal["bigquery"] = "bigquery"
    project_id: str
    dataset_id: str

    @classmethod
    def get_fallback_secret_fields(cls) -> Set[Tuple]:
        """
        The required connection config secrets when namespace metadata is missing.
        For BigQuery, dataset must be provided in secrets if dataset_id is not in namespace metadata.
        """
        schema = secrets_schemas["bigquery"]
        return {("dataset", schema.model_fields["dataset"].title)}



from abc import ABC, abstractmethod
from typing import Any, Dict, Generic, List, Optional

from sqlalchemy.orm import Query, Session

from fides.api.common_exceptions import MessageDispatchException
from fides.api.models.connectionconfig import ConnectionConfig, ConnectionTestStatus
from fides.api.models.privacy_request import PrivacyRequest
from fides.api.models.sql_models import Organization  # type: ignore[attr-defined]
from fides.api.service.connectors.base_connector import DB_CONNECTOR_TYPE
from fides.config import CONFIG


class BaseEmailConnector(Generic[DB_CONNECTOR_TYPE], ABC):
    """
    Abstract BaseEmailConnector that operates at the Dataset Level, not the Collection level.
    """

    @property
    @abstractmethod
    def required_identities(self) -> List[str]:
        """Returns the identity types we need to supply to the third party for this connector."""

    @property
    @abstractmethod
    def identities_for_test_email(self) -> Dict[str, Any]:
        """The mock user identities that are sent in the test email to ensure the connector is working"""

    def __init__(self, configuration: ConnectionConfig):
        self.configuration = configuration
        self.hide_parameters = not CONFIG.dev_mode
        self.db_client: Optional[DB_CONNECTOR_TYPE] = None

    @abstractmethod
    def test_connection(self) -> Optional[ConnectionTestStatus]:
        """
        Used to send a test email to the specified email address to verify the generated email template.
        """

    @abstractmethod
    def batch_email_send(self, privacy_requests: Query) -> None:
        """
        Aggregates the identities provided by multiple privacy requests and sends them in a single batch email.
        """

    @abstractmethod
    def add_skipped_log(self, db: Session, privacy_request: PrivacyRequest) -> None:
        """Add skipped log for the email connector to a privacy request"""


def get_org_name(db: Session) -> str:
    org: Optional[Organization] = (
        db.query(Organization).order_by(Organization.created_at.desc()).first()
    )

    if not org or not org.name:
        raise MessageDispatchException(
            "Cannot send an email to third-party vendor. No organization name found."
        )

    return org.name



"""migrate meta vendor

Revision ID: 671358247251
Revises: ad2e3f9a6850
Create Date: 2023-09-13 20:52:58.425483

"""

import json
from typing import Dict, Optional

import sqlalchemy as sa
from alembic import op
from loguru import logger

# revision identifiers, used by Alembic.
from sqlalchemy import text
from sqlalchemy.engine import Connection, ResultProxy
from sqlalchemy.sql.elements import TextClause

revision = "671358247251"
down_revision = "ad2e3f9a6850"
branch_labels = None
depends_on = None


def upgrade():
    """Migrate system > meta > vendor > id to system > vendor_id if applicable"""
    bind: Connection = op.get_bind()
    existing_systems: ResultProxy = bind.execute(
        text("select id, meta, vendor_id from ctl_systems;")
    )

    for row in existing_systems:
        if row["vendor_id"]:
            logger.warning(
                "Skipped migrating system > meta > vendor > id for system {} as vendor_id already exists",
                row["id"],
            )
            continue

        # Locate system > meta > vendor > id if it exists and is in correct format
        meta: Optional[Dict] = row["meta"] or {}
        if not isinstance(meta, dict):
            continue

        meta_vendor: Optional[Dict] = meta.get("vendor", {})
        meta_vendor_id: Optional[str] = (
            meta_vendor.get("id", None)
            if isinstance(meta_vendor, dict)
            else None  # Just double checking here
        )
        if not meta_vendor_id:
            continue

        # Migrate vendor_id, and remove "vendor" key from "meta" while we're here
        meta.pop("vendor")
        update_vendor_id_query: TextClause = text(
            "UPDATE ctl_systems SET vendor_id= :meta_vendor_id, meta=:updated_meta WHERE id= :id"
        )
        bind.execute(
            update_vendor_id_query,
            {
                "id": row["id"],
                "meta_vendor_id": str(meta_vendor_id),
                "updated_meta": json.dumps(meta),
            },
        )


def downgrade():
    """Migrate system > vendor_id to system > meta > vendor > id if applicable"""
    bind: Connection = op.get_bind()
    existing_systems: ResultProxy = bind.execute(
        text("select id, meta, vendor_id from ctl_systems;")
    )

    for row in existing_systems:
        existing_vendor_id: Optional[str] = row["vendor_id"]
        if not existing_vendor_id:
            continue

        meta: Dict = row["meta"] or {}
        if "vendor" not in meta:
            meta["vendor"] = {}
            meta["vendor"]["id"] = existing_vendor_id

        update_vendor_id_query: TextClause = text(
            "UPDATE ctl_systems SET meta = :updated_meta, vendor_id = null WHERE id= :id"
        )
        bind.execute(
            update_vendor_id_query,
            {"id": row["id"], "updated_meta": json.dumps(meta)},
        )



"""create table for persisting MonitorExecution records

Revision ID: ed96417b07d8
Revises: 1088e8353890
Create Date: 2025-01-27 23:41:59.803285

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "ed96417b07d8"
down_revision = "1088e8353890"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "monitorexecution",
        sa.Column("id", sa.String(), primary_key=True),
        sa.Column("monitor_config_key", sa.String(), nullable=False),
        sa.Column("status", sa.String(), nullable=True),
        sa.Column("started", sa.DateTime(), nullable=True, default=sa.func.now()),
        sa.Column("completed", sa.DateTime(), nullable=True),
        sa.Column(
            "classification_instances",
            postgresql.ARRAY(sa.String()),
            nullable=False,
            default=[],
        ),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        "ix_monitorexecution_monitor_config_key",
        "monitorexecution",
        ["monitor_config_key"],
    )


def downgrade():
    op.drop_table("monitorexecution")
    op.drop_index(
        "ix_monitorexecution_monitor_config_key", table_name="monitorexecution"
    )



"""migrate remaining data categories

Revision ID: a6d9cdfcc7dc
Revises: 31493e48c1d8
Create Date: 2024-07-09 17:51:17.542889

"""

from typing import Dict

from alembic import op
from loguru import logger
from sqlalchemy import text
from sqlalchemy.engine import Connection

from fides.api.alembic.migrations.helpers.fideslang_migration_functions import (
    remove_conflicting_rule_targets,
    update_ctl_policies,
    update_data_label_tables,
    update_datasets_data_categories,
    update_default_dsr_policies,
    update_privacy_declarations,
    update_rule_targets,
    update_system_ingress_egress_data_categories,
)

# revision identifiers, used by Alembic.
revision = "a6d9cdfcc7dc"
down_revision = "f712aa9429f4"
branch_labels = None
depends_on = None

###############
## Data Uses ##
###############
"""
The `key` is the old value, the `value` is the new value
These are ordered specifically so that string replacement works on both parent and child items
"""
data_use_upgrades: Dict[str, str] = {}

#####################
## Data Categories ##
#####################
"""
The `key` is the old value, the `value` is the new value
These are ordered specifically so that string replacement works on both parent and child items
"""
data_category_upgrades: Dict[str, str] = {
    "user.biometric_health": "user.biometric.health",
    "user.credentials.biometric_credentials": "user.authorization.biometric",
    "user.credentials.password": "user.authorization.password",
}


def upgrade() -> None:
    """
    Given that our advice is to turn off auto-migrations and make a db copy,
    there is no "downgrade" version of this. It also wouldn't be feasible given
    it would require an older version of fideslang.
    """
    bind: Connection = op.get_bind()

    logger.info("Removing old default data categories")
    bind.execute(text("DELETE FROM ctl_data_categories WHERE is_default = TRUE;"))

    logger.info("Upgrading additional Privacy Declarations for Fideslang 2.0")
    update_privacy_declarations(bind, data_use_upgrades, data_category_upgrades)

    logger.info("Upgrading additional Policy Rules for Fideslang 2.0")
    update_ctl_policies(bind, data_use_upgrades, data_category_upgrades)

    logger.info("Upgrading additional Data Categories in Datasets")
    update_datasets_data_categories(bind, data_category_upgrades)

    logger.info("Upgrading additional Data Categories in System egress/ingress")
    update_system_ingress_egress_data_categories(bind, data_category_upgrades)

    logger.info("Updating additional Rule Targets")
    update_rule_targets(bind, data_category_upgrades)

    logger.info("Upgrading additional Taxonomy Items for Fideslang 2.0")
    update_data_label_tables(bind, data_category_upgrades, "ctl_data_categories")

    logger.info("Adding new rule targets to default policies")
    update_default_dsr_policies(bind)

    logger.info("Removing conflicting rule targets from all erasure policies")
    remove_conflicting_rule_targets(bind)


def downgrade() -> None:
    """
    This migration does not support downgrades.
    """
    logger.info("Removal of additional Fideslang 2.0 data categories is unsupported.")



"""make datasetconfig.ctl_datasets non-nullable

Revision ID: 9c6f62e4c9da
Revises: 216cdc7944f1
Create Date: 2022-12-09 23:56:13.022119

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "9c6f62e4c9da"
down_revision = "216cdc7944f1"
branch_labels = None
depends_on = None


def upgrade():
    """Followup migration to make datasetconfig.ctl_dataset_id non nullable"""
    op.alter_column(
        "datasetconfig", "ctl_dataset_id", existing_type=sa.VARCHAR(), nullable=False
    )


def downgrade():
    op.alter_column(
        "datasetconfig", "ctl_dataset_id", existing_type=sa.VARCHAR(), nullable=True
    )



"""update system.legal_basis_for_transfers

Revision ID: 3038667ba898
Revises: 507563f6f8d4
Create Date: 2023-08-25 14:19:16.798409

"""

from typing import Dict

import sqlalchemy as sa
from alembic import op
from loguru import logger
from sqlalchemy.orm.session import Session
from sqlalchemy.sql import text

# revision identifiers, used by Alembic.
revision = "3038667ba898"
down_revision = "507563f6f8d4"
branch_labels = None
depends_on = None

legal_basis_transfers_upgrades: Dict[str, str] = {
    "Adequacy decision": "Adequacy Decision",
    "Standard contractual clauses": "SCCs",
    "Binding corporate rules": "BCRs",
}

legal_basis_transfers_downgrades: Dict[str, str] = {
    value: key for key, value in legal_basis_transfers_upgrades.items()
}


def upgrade():
    session = Session(bind=op.get_bind())
    conn = session.connection()
    bind = op.get_bind()

    #############################
    # legal_basis_for_transfers #
    #############################
    statement_transfers = text(
        """SELECT distinct legal_basis_for_transfers, id
FROM ctl_systems"""
    )
    existing_values_transfers = conn.execute(statement_transfers)

    # drop column since we will repopulating later
    op.drop_column(
        "ctl_systems",
        "legal_basis_for_transfers",
    )

    # add column back, but make it multi-value
    op.add_column(
        "ctl_systems",
        sa.Column(
            "legal_basis_for_transfers",
            type_=sa.ARRAY(sa.String()),
            server_default="{}",
            nullable=True,
        ),
    )

    # data migration - updates to some values since allowable values enum has changed
    # (enum is not enforced/specified on the DB, just API-level)

    for row in existing_values_transfers:
        existing_value = row[0]
        updated_value = legal_basis_transfers_upgrades.get(
            existing_value, existing_value
        )
        updated_value = (
            [updated_value] if updated_value else []
        )  # wrap in a list because column is now array type
        bind.execute(
            text(
                "UPDATE ctl_systems SET legal_basis_for_transfers = :updated_basis WHERE id = :id"
            ),
            {"updated_basis": updated_value, "id": row[1]},
        )

    #############################
    # legal_basis_for_profiling #
    #############################
    statement_profiling = text(
        """SELECT distinct legal_basis_for_profiling, id
FROM ctl_systems"""
    )
    existing_values_profiling = conn.execute(statement_profiling)

    op.drop_column(
        "ctl_systems",
        "legal_basis_for_profiling",
    )
    op.add_column(
        "ctl_systems",
        sa.Column(
            "legal_basis_for_profiling",
            type_=sa.ARRAY(sa.String()),
            server_default="{}",
            nullable=True,
        ),
    )

    # data migration - wrap values in a list
    for row in existing_values_profiling:
        existing_value = row[0]
        updated_value = legal_basis_transfers_upgrades.get(
            existing_value, existing_value
        )
        updated_value = (
            [updated_value] if updated_value else []
        )  # wrap in a list because column is now array type
        bind.execute(
            text(
                "UPDATE ctl_systems SET legal_basis_for_profiling = :updated_basis WHERE id = :id"
            ),
            {"updated_basis": updated_value, "id": row[1]},
        )


def downgrade():
    """
    Downgrade from array fields to single-value.

    We do not try to perform a data migration, since we'd likely have to
    drop some values arbitrarily. Better to just null the column out
    to make it clear that manual intervention is needed.
    """
    logger.warning(
        "Database migration downgrade will remove all values for `system.legal_basis_for_transfers` and `system.legal_basis_for_profiling` fields. \
        Values for these fields must be repopulated manually!"
    )
    op.drop_column(
        "ctl_systems",
        "legal_basis_for_transfers",
    )
    op.drop_column(
        "ctl_systems",
        "legal_basis_for_profiling",
    )

    op.add_column(
        "ctl_systems",
        sa.Column(
            "legal_basis_for_transfers",
            type_=sa.String(),
            nullable=True,
        ),
    )
    op.add_column(
        "ctl_systems",
        sa.Column(
            "legal_basis_for_profiling",
            type_=sa.String(),
            nullable=True,
        ),
    )



"""Remove qualifier lists from data set models

Revision ID: 45c7a349db68
Revises: 732105cd54e3
Create Date: 2021-10-25 17:59:25.244689

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "45c7a349db68"
down_revision = "732105cd54e3"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("datasets", sa.Column("data_qualifier", sa.String(), nullable=True))
    op.drop_column("datasets", "data_qualifiers")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "datasets",
        sa.Column(
            "data_qualifiers",
            postgresql.ARRAY(sa.VARCHAR()),
            autoincrement=False,
            nullable=True,
        ),
    )
    op.drop_column("datasets", "data_qualifier")
    # ### end Alembic commands ###



# pylint: disable=too-many-lines
import re
from abc import ABC, abstractmethod
from typing import Any, Dict, Generic, List, Optional, Tuple, Type, TypeVar

import pydash
from loguru import logger
from pydantic import ValidationError
from sqlalchemy import text
from sqlalchemy.sql import Executable  # type: ignore
from sqlalchemy.sql.elements import TextClause

from fides.api.common_exceptions import MissingNamespaceSchemaException
from fides.api.graph.config import (
    ROOT_COLLECTION_ADDRESS,
    CollectionAddress,
    Field,
    FieldPath,
    MaskingTruncation,
)
from fides.api.graph.execution import ExecutionNode
from fides.api.models.policy import Policy, Rule
from fides.api.models.privacy_request import PrivacyRequest
from fides.api.schemas.namespace_meta.namespace_meta import NamespaceMeta
from fides.api.schemas.policy import ActionType
from fides.api.service.masking.strategy.masking_strategy import MaskingStrategy
from fides.api.service.masking.strategy.masking_strategy_nullify import (
    NullMaskingStrategy,
)
from fides.api.task.refine_target_path import (
    build_refined_target_paths,
    join_detailed_path,
)
from fides.api.util.collection_util import Row, append, filter_nonempty_values
from fides.api.util.logger import Pii
from fides.api.util.querytoken import QueryToken

T = TypeVar("T")


class QueryConfig(Generic[T], ABC):
    """A wrapper around a resource-type dependent query object that can generate runnable queries
    and string representations."""

    def __init__(self, node: ExecutionNode):
        self.node = node

    @property
    def partitioning(self) -> Optional[Dict]:  # pylint: disable=R1711
        # decided to de-scope partitioning support to only bigquery as this grew more complex,
        # but keeping more generic support stubbed out feels like a reasonable step.
        if self.node.collection.partitioning:
            logger.warning(
                "Partitioning is only supported on BigQuery connectors at this time!"
            )
        return None

    def field_map(self) -> Dict[FieldPath, Field]:
        """Flattened FieldPaths of interest from this traversal_node."""
        return self.node.collection.field_dict

    def top_level_field_map(self) -> Dict[FieldPath, Field]:
        """Top level FieldPaths on this traversal_node."""
        return self.node.collection.top_level_field_dict

    def build_rule_target_field_paths(
        self, policy: Policy
    ) -> Dict[Rule, List[FieldPath]]:
        """
        Return dictionary of rules mapped to update-able field paths on a given collection
        Example:
        {<fides.api.models.policy.Rule object at 0xffff9160e190>: [FieldPath('name'), FieldPath('code'), FieldPath('ccn')]}
        """
        rule_updates: Dict[Rule, List[FieldPath]] = {}
        for rule in policy.rules:  # type: ignore[attr-defined]
            if rule.action_type != ActionType.erasure:
                continue
            rule_categories: List[str] = rule.get_target_data_categories()
            if not rule_categories:
                continue

            targeted_field_paths = []
            collection_categories: Dict[
                str, List[FieldPath]
            ] = self.node.collection.field_paths_by_category  # type: ignore
            for rule_cat in rule_categories:
                for collection_cat, field_paths in collection_categories.items():
                    if collection_cat.startswith(rule_cat):
                        targeted_field_paths.extend(field_paths)
            rule_updates[rule] = targeted_field_paths

        return rule_updates

    @property
    def primary_key_field_paths(self) -> Dict[FieldPath, Field]:
        """Mapping of FieldPaths to Fields that are marked as PK's"""
        return {
            field_path: field
            for field_path, field in self.field_map().items()
            if field.primary_key
        }

    @property
    def reference_field_paths(self) -> Dict[FieldPath, Field]:
        """Mapping of FieldPaths to Fields that have incoming identity or dataset references"""
        return {
            field_path: field
            for field_path, field in self.field_map().items()
            if field_path in {edge.f2.field_path for edge in self.node.incoming_edges}
        }

    def query_sources(self) -> Dict[str, List[CollectionAddress]]:
        """Display the input collection(s) for each query key for display purposes.

        {'user_info.user_id': [postgres_db:users]}

        Translate keys from field paths to string values
        """
        data: Dict[str, List[CollectionAddress]] = {}
        for edge in self.node.incoming_edges:
            append(data, edge.f2.field_path.string_path, edge.f1.collection_address())
        return data

    def display_query_data(self) -> Dict[str, Any]:
        """Data to represent a display (dry-run) query. Since we don't know
        what data is available, just generate a query where the input identity
        values are assumed to be present and singular and all other values that
        may be multiple are represented by a pair [?,?]"""

        data = {}
        t = QueryToken()

        for field_str, input_collection_address in self.query_sources().items():
            if (
                len(input_collection_address) == 1
                and input_collection_address[0] == ROOT_COLLECTION_ADDRESS
            ):
                data[field_str] = [t]
            else:
                data[field_str] = [
                    t,
                    QueryToken(),
                ]  # intentionally want a second instance so that set does not collapse into 1 value

        return data

    def _is_child_of_masked_parent(
        self, path: str, masked_parent_paths: List[str]
    ) -> bool:
        """
        Check if the given path is a child of any already masked parent object.
        """
        for parent_path in masked_parent_paths:
            if path == parent_path or path.startswith(parent_path + "."):
                return True
        return False

    def _is_object_or_array_with_data_category(
        self, field_val: Any, field: Optional[Field]
    ) -> bool:
        """
        Check if field is an object or array of objects with data categories
        """
        # First check if the field has data categories
        has_data_category = (
            field and field.data_categories and len(field.data_categories) > 0
        )

        if not has_data_category:
            return False

        # Check if it's an object
        is_object = isinstance(field_val, dict)

        # Check if it's a non-empty array of objects
        is_array_of_objects = False
        if (
            isinstance(field_val, list) and field_val
        ):  # Check if it's a list and not empty
            # Only check first element if the list is not empty
            is_array_of_objects = isinstance(field_val[0], dict)

        # Return true if it's either an object or an array of objects and has data categories
        return is_object or is_array_of_objects

    def update_value_map(  # pylint: disable=R0914
        self, row: Row, policy: Policy, request: PrivacyRequest
    ) -> Dict[str, Any]:
        """Map the relevant field (as strings) to be updated on the row with their masked values from Policy Rules

        Example return:  {'name': None, 'ccn': None, 'code': None, 'workplace_info.employer': None, 'children.0': None}

        In this example, a Null Masking Strategy was used to determine that the name/ccn/code fields, nested
        workplace_info.employer field, and the first element in 'children' for a given customer_id will be replaced
        with null values.
        """
        rule_to_collection_field_paths: Dict[Rule, List[FieldPath]] = (
            self.build_rule_target_field_paths(policy)
        )

        value_map: Dict[str, Any] = {}
        # Track which parent paths have been masked as whole objects/arrays
        masked_parent_paths: List[str] = []

        for rule, field_paths in rule_to_collection_field_paths.items():
            strategy_config = rule.masking_strategy
            if not strategy_config:
                continue

            for rule_field_path in field_paths:
                # Check if field is read-only before processing
                field = self.field_map().get(rule_field_path)
                if field and field.read_only:
                    logger.debug(
                        f"Skipping read-only field: {rule_field_path.string_path}"
                    )
                    continue

                # Skip if this field is a child of an already masked parent object
                if self._is_child_of_masked_parent(
                    rule_field_path.string_path, masked_parent_paths
                ):
                    logger.debug(
                        f"Skipping field {rule_field_path.string_path} because its parent object has already been masked"
                    )
                    continue

                # Determine masking strategy
                strategy: MaskingStrategy = MaskingStrategy.get_strategy(
                    strategy_config["strategy"], strategy_config["configuration"]
                )

                # Apply field-level masking strategy override if present
                if field and field.masking_strategy_override:
                    masking_strategy_override = field.masking_strategy_override
                    strategy = MaskingStrategy.get_strategy(
                        masking_strategy_override.strategy,
                        masking_strategy_override.configuration,  # type: ignore[arg-type]
                    )
                    logger.warning(
                        f"Using field-level masking override of type '{strategy.name}' for {rule_field_path.string_path}"
                    )

                null_masking: bool = strategy.name == NullMaskingStrategy.name
                truncation: MaskingTruncation = [
                    MaskingTruncation(field.data_type_converter, field.length)
                    for field_path, field in self.field_map().items()
                    if field_path == rule_field_path
                ][0]

                if not self._supported_data_type(truncation, null_masking, strategy):
                    logger.warning(
                        "Unable to generate a query for field {}: data_type is either not present on the field or not supported for the {} masking strategy. Received data type: {}",
                        rule_field_path.string_path,
                        strategy.name,
                        truncation.data_type_converter.name,  # type: ignore
                    )
                    continue

                # Get the actual field value to determine its type
                field_val = pydash.objects.get(row, rule_field_path.string_path)

                # Handle objects and arrays with data categories - mask as whole entities
                if self._is_object_or_array_with_data_category(field_val, field):
                    logger.info(
                        f"Field {rule_field_path.string_path} is an object or array of objects with data category. Masking entire field."
                    )
                    if field_val is not None:
                        value_map[rule_field_path.string_path] = (
                            self._generate_masked_value(
                                request_id=request.id,
                                strategy=strategy,
                                val=field_val,
                                masking_truncation=truncation,
                                null_masking=null_masking,
                                str_field_path=rule_field_path.string_path,
                            )
                        )
                        # Add this path to masked parent paths to skip its children later
                        masked_parent_paths.append(rule_field_path.string_path)
                else:
                    # Standard approach: build refined target paths for individual fields
                    paths_to_mask = [
                        join_detailed_path(path)
                        for path in build_refined_target_paths(
                            row, query_paths={rule_field_path: None}
                        )
                    ]

                    # Process each detailed path, skipping those that are children of masked parents
                    for detailed_path in paths_to_mask:
                        if self._is_child_of_masked_parent(
                            detailed_path, masked_parent_paths
                        ):
                            logger.debug(
                                f"Skipping detailed path {detailed_path} because its parent object has already been masked"
                            )
                            continue

                        value_map[detailed_path] = self._generate_masked_value(
                            request_id=request.id,
                            strategy=strategy,
                            val=pydash.objects.get(row, detailed_path),
                            masking_truncation=truncation,
                            null_masking=null_masking,
                            str_field_path=detailed_path,
                        )

        return value_map

    @staticmethod
    def _supported_data_type(
        masking_truncation: MaskingTruncation,
        null_masking: bool,
        strategy: MaskingStrategy,
    ) -> bool:
        """Helper method to determine whether given data_type exists and is supported by the masking strategy"""
        if null_masking:
            return True
        if not masking_truncation.data_type_converter:
            return False
        if not strategy.data_type_supported(
            data_type=masking_truncation.data_type_converter.name
        ):
            return False
        return True

    @staticmethod
    def _generate_masked_value(  # pylint: disable=R0913
        request_id: str,
        strategy: MaskingStrategy,
        val: Any,
        masking_truncation: MaskingTruncation,
        null_masking: bool,
        str_field_path: str,
    ) -> T:
        # masking API takes and returns lists, but here we are only leveraging single elements
        masked_val = strategy.mask([val], request_id)[0]  # type: ignore

        logger.debug(
            "Generated the following masked val for field {}: {}",
            str_field_path,
            masked_val,
        )

        # special case for null masking
        if null_masking:
            return masked_val

        if masking_truncation.length:
            logger.warning(
                "Because a length has been specified for field {}, we will truncate length of masked value to match, regardless of masking strategy",
                str_field_path,
            )
            #  for strategies other than null masking we assume that masked data type is the same as specified data type
            masked_val = masking_truncation.data_type_converter.truncate(  # type: ignore
                masking_truncation.length, masked_val
            )
        return masked_val

    @abstractmethod
    def generate_query(
        self, input_data: Dict[str, List[Any]], policy: Optional[Policy]
    ) -> Optional[T]:
        """Generate a retrieval query. If there is no data to be queried
        (for example, if the policy identifies no fields to be queried)
        returns None"""

    @abstractmethod
    def query_to_str(self, t: T, input_data: Dict[str, List[Any]]) -> Optional[str]:
        """Convert query to string"""

    @abstractmethod
    def dry_run_query(self) -> Optional[str]:
        """dry run query for display"""

    @abstractmethod
    def generate_update_stmt(
        self, row: Row, policy: Policy, request: PrivacyRequest
    ) -> Optional[T]:
        """Generate an update statement. If there is no data to be updated
        (for example, if the policy identifies no fields to be updated)
        returns None"""


class SQLLikeQueryConfig(QueryConfig[T], ABC):
    """
    Abstract query config for SQL-like languages (that may not be strictly SQL).
    """

    namespace_meta_schema: Optional[Type[NamespaceMeta]] = None

    def __init__(self, node: ExecutionNode, namespace_meta: Optional[Dict] = None):
        super().__init__(node)
        self.namespace_meta: Optional[NamespaceMeta] = None

        if namespace_meta is not None:
            if self.namespace_meta_schema is None:
                raise MissingNamespaceSchemaException(
                    f"{self.__class__.__name__} must define a namespace_meta_schema when namespace_meta is provided."
                )
            try:
                self.namespace_meta = self.namespace_meta_schema.model_validate(
                    namespace_meta
                )
            except ValidationError as exc:
                raise ValueError(f"Invalid namespace_meta: {exc}")

    def format_fields_for_query(
        self,
        field_paths: List[FieldPath],
    ) -> List[str]:
        """Returns field paths in a format they can be added into SQL queries.

        This currently takes no nesting into account and only returns the
        last value from each key. It will need to be updated to support
        nested values.
        """
        return [fk.levels[-1] for fk in field_paths]

    @abstractmethod
    def format_query_data_name(self, query_data_name: str) -> str:
        """Returns query_data_name formatted according to specific SQL dialect"""

    @abstractmethod
    def get_formatted_query_string(
        self,
        field_list: str,
        clauses: List[str],
    ) -> str:
        """Returns a SQL query string."""

    @abstractmethod
    def format_clause_for_query(
        self, string_path: str, operator: str, operand: str
    ) -> str:
        """Returns clause formatted in the specific SQL dialect for the query"""

    def generate_raw_query_without_tuples(
        self, field_list: List[str], filters: Dict[str, List[Any]]
    ) -> Optional[T]:
        """
        Generate a raw query using the provided field_list and filters,
        i.e a query of the form:
        SELECT <field_list> FROM <collection> WHERE <filters>

        Generates distinct key/val pairs for building the query string instead of a tuple.

        E.g. SQLQueryConfig uses 1 key as a tuple:
        SELECT order_id,product_id,quantity FROM order_item WHERE order_id IN (:some-params-in-tuple)
        which for some connectors gets interpreted as data types (mssql) or quotes (bigquery).

        This method produces distinct keys for the query_str:
        SELECT order_id,product_id,quantity FROM order_item WHERE order_id IN (:_in_stmt_generated_0, :_in_stmt_generated_1, :_in_stmt_generated_2)

        """
        clauses = []
        query_data: Dict[str, Tuple[Any, ...]] = {}
        for field_name, field_value in filters.items():
            data = set(field_value)
            if len(data) == 1:
                clauses.append(
                    self.format_clause_for_query(field_name, "=", field_name)
                )
                query_data[field_name] = data.pop()
            elif len(data) > 1:
                data_vals = list(data)
                query_data_keys: List[str] = []
                for val in data_vals:
                    # appending "_in_stmt_generated_" (can be any arbitrary str) so that this name has less change of conflicting with pre-existing column in table
                    query_data_name = (
                        field_name + "_in_stmt_generated_" + str(data_vals.index(val))
                    )
                    query_data[query_data_name] = val
                    query_data_keys.append(self.format_query_data_name(query_data_name))
                operand = ", ".join(query_data_keys)
                clauses.append(self.format_clause_for_query(field_name, "IN", operand))

        if len(clauses) > 0:
            formatted_fields = ", ".join(field_list)
            query_str = self.get_formatted_query_string(formatted_fields, clauses)
            return self.format_query_stmt(query_str, query_data)

        return None

    def generate_query_without_tuples(  # pylint: disable=R0914
        self,
        input_data: Dict[str, List[Any]],
        policy: Optional[Policy] = None,
    ) -> Optional[T]:
        """
        Generate a retrieval query. Generates distinct key/val pairs for building the query string instead of a tuple.

        E.g. SQLQueryConfig uses 1 key as a tuple:
        SELECT order_id,product_id,quantity FROM order_item WHERE order_id IN (:some-params-in-tuple)
        which for some connectors gets interpreted as data types (mssql) or quotes (bigquery).

        This override produces distinct keys for the query_str:
        SELECT order_id,product_id,quantity FROM order_item WHERE order_id IN (:_in_stmt_generated_0, :_in_stmt_generated_1, :_in_stmt_generated_2)
        """
        filtered_data = self.node.typed_filtered_values(input_data)

        if filtered_data:
            formatted_fields = self.format_fields_for_query(
                list(self.field_map().keys())
            )

            return self.generate_raw_query_without_tuples(
                formatted_fields, filtered_data
            )

        logger.warning(
            "There is not enough data to generate a valid query for {}",
            self.node.address,
        )
        return None

    def get_update_stmt(
        self,
        update_clauses: List[str],
        where_clauses: List[str],
    ) -> str:
        """Returns a SQL UPDATE statement to fit SQL syntax."""
        return f"UPDATE {self.node.address.collection} SET {', '.join(update_clauses)} WHERE {' AND '.join(where_clauses)}"

    @abstractmethod
    def get_update_clauses(
        self,
        update_value_map: Dict[str, Any],
        where_clause_fields: Dict[str, Field],
    ) -> List[str]:
        """Returns a list of update clauses for the update statement."""

    @abstractmethod
    def format_query_stmt(self, query_str: str, update_value_map: Dict[str, Any]) -> T:
        """Returns a formatted update statement in the appropriate dialect."""

    @abstractmethod
    def format_key_map_for_update_stmt(self, param_map: Dict[str, Any]) -> List[str]:
        """Adds the appropriate formatting for update statements in this datastore."""

    def generate_update_stmt(
        self, row: Row, policy: Policy, request: PrivacyRequest
    ) -> Optional[T]:
        """Returns an update statement in generic SQL-ish dialect."""
        update_value_map: Dict[str, Any] = self.update_value_map(row, policy, request)

        non_empty_primary_key_fields: Dict[str, Field] = filter_nonempty_values(
            {
                fpath.string_path: fld.cast(row[fpath.string_path])
                for fpath, fld in self.primary_key_field_paths.items()
                if fpath.string_path in row
            }
        )

        non_empty_reference_fields: Dict[str, Field] = filter_nonempty_values(
            {
                fpath.string_path: fld.cast(row[fpath.string_path])
                for fpath, fld in self.reference_field_paths.items()
                if fpath.string_path in row
            }
        )

        # Create parameter mappings with masked_ prefix for SET values
        param_map = {
            **{f"masked_{k}": v for k, v in update_value_map.items()},
            **non_empty_primary_key_fields,
            **non_empty_reference_fields,
        }

        update_clauses = self.get_update_clauses(
            {k: f"masked_{k}" for k in update_value_map},
            non_empty_primary_key_fields or non_empty_reference_fields,
        )
        where_clauses = self.format_key_map_for_update_stmt(
            {k: k for k in non_empty_primary_key_fields or non_empty_reference_fields}
        )

        valid = len(where_clauses) > 0 and len(update_clauses) > 0
        if not valid:
            logger.warning(
                "There is not enough data to generate a valid update statement for {}",
                self.node.address,
            )
            return None

        query_str = self.get_update_stmt(update_clauses, where_clauses)
        logger.info("query = {}, params = {}", Pii(query_str), Pii(param_map))
        return self.format_query_stmt(query_str, param_map)


class SQLQueryConfig(SQLLikeQueryConfig[Executable]):
    """Query config that translates parameters into SQL statements."""

    def generate_raw_query(
        self, field_list: List[str], filters: Dict[str, List[Any]]
    ) -> Optional[TextClause]:
        """
        Generate a raw query using the provided field_list and filters,
        i.e a query of the form:
        SELECT <field_list> FROM <collection> WHERE <filters>
        """
        clauses = []
        query_data: Dict[str, Tuple[Any, ...]] = {}
        for field_name, field_value in filters.items():
            data = set(field_value)
            if len(data) == 1:
                clauses.append(
                    self.format_clause_for_query(field_name, "=", field_name)
                )
                query_data[field_name] = (data.pop(),)
            elif len(data) > 1:
                clauses.append(
                    self.format_clause_for_query(field_name, "IN", field_name)
                )
                query_data[field_name] = tuple(data)

        if len(clauses) > 0:
            formatted_fields = ", ".join(field_list)
            query_str = self.get_formatted_query_string(formatted_fields, clauses)
            return text(query_str).params(query_data)

        return None

    def format_clause_for_query(
        self, string_path: str, operator: str, operand: str
    ) -> str:
        """Returns clauses in a format they can be added into SQL queries."""
        return f"{string_path} {operator} :{operand}"

    def get_formatted_query_string(
        self,
        field_list: str,
        clauses: List[str],
    ) -> str:
        """Returns an SQL query string."""
        return f"SELECT {field_list} FROM {self.node.collection.name} WHERE {' OR '.join(clauses)}"

    def generate_query(
        self,
        input_data: Dict[str, List[Any]],
        policy: Optional[Policy] = None,
    ) -> Optional[TextClause]:
        """Generate a retrieval query"""
        filtered_data: Dict[str, Any] = self.node.typed_filtered_values(input_data)

        if filtered_data:
            formatted_fields: List[str] = self.format_fields_for_query(
                list(self.field_map().keys())
            )

            return self.generate_raw_query(formatted_fields, filtered_data)

        logger.warning(
            "There is not enough data to generate a valid query for {}",
            self.node.address,
        )
        return None

    def format_key_map_for_update_stmt(self, param_map: Dict[str, Any]) -> List[str]:
        """Adds the appropriate formatting for update statements in this datastore."""
        return [f"{k} = :{v}" for k, v in sorted(param_map.items())]

    def get_update_clauses(
        self,
        update_value_map: Dict[str, Any],
        where_clause_fields: Dict[str, Field],
    ) -> List[str]:
        """Returns a list of update clauses for the update statement."""
        return self.format_key_map_for_update_stmt(update_value_map)

    def format_query_stmt(
        self, query_str: str, update_value_map: Dict[str, Any]
    ) -> Executable:
        """Returns a formatted update statement in the appropriate dialect."""
        return text(query_str).params(update_value_map)

    def query_to_str(self, t: TextClause, input_data: Dict[str, List[Any]]) -> str:
        """string representation of a query for logging/dry-run"""

        def transform_param(p: Any) -> str:
            if isinstance(p, str):
                return f"'{p}'"
            return str(p)

        query_str = str(t)
        for k, v in input_data.items():
            if len(v) == 1:
                query_str = re.sub(f"= :{k}", f"= {transform_param(v[0])}", query_str)
            elif len(v) > 0:
                query_str = re.sub(f"IN :{k}", f"IN { tuple(set(v)) }", query_str)
        return query_str

    def dry_run_query(self) -> Optional[str]:
        """Returns a text representation of the query."""
        query_data = self.display_query_data()
        text_clause = self.generate_query(query_data, None)
        if text_clause is not None:
            return self.query_to_str(text_clause, query_data)
        return None

    def format_query_data_name(self, query_data_name: str) -> str:
        return f":{query_data_name}"


class QueryStringWithoutTuplesOverrideQueryConfig(SQLQueryConfig):
    """
    Generates SQL valid for connectors that require the query string to be built without tuples.
    """

    # Overrides SQLQueryConfig.generate_raw_query
    def generate_raw_query(
        self, field_list: List[str], filters: Dict[str, List[Any]]
    ) -> Optional[TextClause]:
        """
        Allows executing a somewhat raw query where the field_list and filters do not depend
        on the Node or Graph structure.
        """
        clauses = []
        query_data = {}
        for field_name, field_value in filters.items():
            data = set(field_value)
            if len(data) == 1:
                clauses.append(
                    self.format_clause_for_query(field_name, "=", field_name)
                )
                query_data[field_name] = data.pop()
            elif len(data) > 1:
                data_vals = list(data)
                query_data_keys: List[str] = []
                for val in data_vals:
                    # appending "_in_stmt_generated_" (can be any arbitrary str) so that this name has lower chance of conflicting with pre-existing column in table
                    query_data_name = (
                        field_name + "_in_stmt_generated_" + str(data_vals.index(val))
                    )
                    query_data[query_data_name] = val
                    query_data_keys.append(self.format_query_data_name(query_data_name))
                operand = ", ".join(query_data_keys)
                clauses.append(self.format_clause_for_query(field_name, "IN", operand))

        if len(clauses) > 0:
            formatted_fields = ", ".join(field_list)
            query_str = self.get_formatted_query_string(formatted_fields, clauses)
            return text(query_str).params(query_data)

        return None

    # Overrides SQLConnector.format_clause_for_query
    def format_clause_for_query(
        self, string_path: str, operator: str, operand: str
    ) -> str:
        """
        Returns clauses in a format they can be added into SQL queries.
        Expects the operand of IN statements to be formatted in the following manner so that
        these distinct keys can be appended to the clause:
        ":_in_stmt_generated_0, :_in_stmt_generated_1, :_in_stmt_generated_2"
        """
        if operator == "IN":
            return f"{string_path} IN ({operand})"
        return super().format_clause_for_query(string_path, operator, operand)

    # Overrides SQLConnector.generate_query
    def generate_query(  # pylint: disable=R0914
        self,
        input_data: Dict[str, List[Any]],
        policy: Optional[Policy] = None,
    ) -> Optional[TextClause]:
        """
        Generate a retrieval query. Generates distinct key/val pairs for building the query string instead of a tuple.

        E.g. The base SQLQueryConfig uses 1 key as a tuple:
        SELECT order_id,product_id,quantity FROM order_item WHERE order_id IN (:some-params-in-tuple)
        which for some connectors gets interpreted as data types (mssql) or quotes (bigquery).

        This override produces distinct keys for the query_str:
        SELECT order_id,product_id,quantity FROM order_item WHERE order_id IN (:_in_stmt_generated_0, :_in_stmt_generated_1, :_in_stmt_generated_2)
        """
        return self.generate_query_without_tuples(input_data, policy)



from typing import List, Optional

from fideslang.models import Dataset
from pydantic import BaseModel, field_validator

from fides.api.models.datasetconfig import validate_masking_strategy_override
from fides.api.schemas.policy import ActionType
from fides.api.schemas.saas.saas_config import SaaSConfig
from fides.api.util.saas_util import load_config_from_string, load_dataset_from_string


class ConnectorTemplate(BaseModel):
    """
    A collection of artifacts that make up a complete
    SaaS connector (SaaS config, dataset, icon, etc.)
    """

    config: str
    dataset: str
    icon: Optional[str] = None
    human_readable: str
    authorization_required: bool
    user_guide: Optional[str] = None
    supported_actions: List[ActionType]

    @field_validator("config")
    @classmethod
    def validate_config(cls, value: str) -> str:
        """Validates the config at the given path"""
        saas_config = SaaSConfig(**load_config_from_string(value))
        if saas_config.fides_key != "<instance_fides_key>":
            raise ValueError(
                "Hard-coded fides_key detected in the config, replace all instances of it with <instance_fides_key>"
            )
        return value

    @field_validator("dataset")
    @classmethod
    def validate_dataset(cls, dataset: str) -> str:
        """Validates the dataset at the given path"""
        saas_dataset = Dataset(**load_dataset_from_string(dataset))
        validate_masking_strategy_override(saas_dataset)
        if saas_dataset.fides_key != "<instance_fides_key>":
            raise ValueError(
                "Hard-coded fides_key detected in the dataset, replace all instances of it with <instance_fides_key>"
            )
        return dataset



import json
import os
import time as time_module
import zipfile
from collections import defaultdict
from io import BytesIO
from pathlib import Path
from typing import Any, Optional

import jinja2
from jinja2 import Environment, FileSystemLoader
from loguru import logger

from fides.api.models.privacy_request import PrivacyRequest
from fides.api.schemas.policy import ActionType
from fides.api.util.storage_util import StorageJSONEncoder

DSR_DIRECTORY = Path(__file__).parent.resolve()

TEXT_COLOR = "#4A5568"
HEADER_COLOR = "#FAFAFA"
BORDER_COLOR = "#E2E8F0"


# pylint: disable=too-many-instance-attributes
class DsrReportBuilder:
    """
    Manages populating HTML templates from the given data and adding the generated
    pages to a zip file in a way that the pages can be navigated between.

    The zip file is structured as follows:
    - welcome.html: the main index page
    - data/dataset_name/index.html: the index page for the dataset
    - data/dataset_name/collection_name/index.html: the index page for the collection
    - data/dataset_name/collection_name/item_index.html: the detail page for the item
    - attachments/index.html: the index page for the attachments

    Args:
        privacy_request: the privacy request object
        dsr_data: the DSR data
    """

    def __init__(
        self,
        privacy_request: PrivacyRequest,
        dsr_data: dict[str, Any],
    ):
        """
        Initializes the DSR report builder.
        """
        # Define pretty_print function for Jinja templates
        jinja2.filters.FILTERS["pretty_print"] = lambda value, indent=4: json.dumps(
            value, indent=indent, cls=StorageJSONEncoder
        )

        # Initialize instance zip file variables
        self.baos = BytesIO()

        # we close this in the finally block of generate()
        # pylint: disable=consider-using-with
        self.out = zipfile.ZipFile(self.baos, "w")
        self.template_loader = Environment(
            loader=FileSystemLoader(DSR_DIRECTORY), autoescape=True
        )

        # to pass in custom colors in the future
        self.template_data: dict[str, Any] = {
            "text_color": TEXT_COLOR,
            "header_color": HEADER_COLOR,
            "border_color": BORDER_COLOR,
        }
        self.main_links: dict[str, Any] = {}  # used to track the generated pages

        # report data to populate the templates
        self.request_data = _map_privacy_request(privacy_request)
        self.dsr_data = dsr_data

        # Track used filenames across all attachments
        self.used_filenames: set[str] = set()

    def _populate_template(
        self,
        template_path: str,
        heading: Optional[str] = None,
        description: Optional[str] = None,
        data: Optional[dict[str, Any]] = None,
    ) -> str:
        """
        Populates the template with the given data.

        Args:
            template_path: the path to the template to populate
            heading: the heading to display on the template
            description: the description to display on the template
            data: the data to populate the template with

        Returns:
            The rendered template as a string.
        """
        report_data = {
            "heading": heading,
            "description": description,
            "data": data,
            "request": self.request_data,
        }
        report_data.update(self.template_data)
        template = self.template_loader.get_template(template_path)
        rendered_template = template.render(report_data)
        return rendered_template

    def _add_file(self, filename: str, contents: str) -> None:
        """
        Adds a file to the zip file.

        Args:
            filename: the name of the file to add
            contents: the contents of the file to add
        """
        if filename and contents:
            self.out.writestr(f"{filename}", contents.encode("utf-8"))

    def _add_dataset(self, dataset_name: str, collections: dict[str, Any]) -> None:
        """
        Generates a page for each collection in the dataset and an index page for the dataset.
        Tracks the generated links to build a root level index after each collection has been processed.

        Args:
            dataset_name: the name of the dataset to add
            collections: the collections to add to the dataset
        """
        # track links to collection indexes
        collection_links = {}
        for collection_name, rows in collections.items():
            collection_url = f"{collection_name}/index.html"
            self._add_collection(rows, dataset_name, collection_name)
            collection_links[collection_name] = collection_url

        # generate dataset index page
        self._add_file(
            f"data/{dataset_name}/index.html",
            self._populate_template(
                "templates/dataset_index.html",
                dataset_name,
                None,
                collection_links,
            ),
        )

    def _get_unique_filename(self, filename: str) -> str:
        """
        Generates a unique filename by appending a counter if the file already exists.
        Now tracks filenames across all directories to ensure global uniqueness.

        Args:
            filename: The original filename

        Returns:
            A unique filename that won't conflict with existing files
        """
        base_name, extension = os.path.splitext(filename)
        counter = 1
        unique_filename = filename

        # Check if file exists in used_filenames set
        while unique_filename in self.used_filenames:
            unique_filename = f"{base_name}_{counter}{extension}"
            counter += 1

        # Add the new filename to the set
        self.used_filenames.add(unique_filename)
        return unique_filename

    def _write_attachment_content(
        self,
        attachments: list[dict[str, Any]],
        directory: str,
    ) -> dict[str, dict[str, str]]:
        """
        Processes attachments and returns a dictionary mapping filenames to their download URLs and sizes.

        Args:
            attachments: The attachments to process
            directory: The directory path (unused for presigned URLs)

        Returns:
            Dictionary mapping filenames to dictionaries containing url and size
        """
        # First process all attachments into a list of tuples (filename, data)
        processed_attachments = []

        for attachment in attachments:
            if not isinstance(attachment, dict):
                continue

            file_name = attachment.get("file_name")
            if not file_name:
                logger.warning("Skipping attachment with no file name")
                continue

            download_url = attachment.get("download_url")
            if not download_url:
                logger.warning("Skipping attachment with no download URL")
                continue

            file_size = attachment.get("file_size")
            if isinstance(file_size, (int, float)):
                file_size = self._format_size(float(file_size))
            else:
                file_size = "Unknown"

            # Get a unique filename to prevent duplicates
            unique_filename = self._get_unique_filename(file_name)

            # Add to processed attachments
            processed_attachments.append(
                (unique_filename, {"url": download_url, "size": file_size})
            )

        # Convert list of tuples to dictionary
        return dict(processed_attachments)

    def _add_collection(
        self, rows: list[dict[str, Any]], dataset_name: str, collection_name: str
    ) -> None:
        """
        Adds a collection to the zip file.

        Args:
            rows: the rows to add to the collection
            dataset_name: the name of the dataset to add the collection to
            collection_name: the name of the collection to add
        """
        items_content = []

        for index, collection_item in enumerate(rows, 1):
            # Create a copy of the item data to avoid modifying the original
            item_data = collection_item.copy()

            # Process any attachments in the item
            if "attachments" in item_data and isinstance(
                item_data["attachments"], list
            ):
                # Process attachments and get their URLs
                attachment_links = self._write_attachment_content(
                    item_data["attachments"],
                    f"data/{dataset_name}/{collection_name}",
                )
                # Add the attachment URLs to the item data
                item_data["attachments"] = attachment_links

            # Add item content to the list
            items_content.append(
                {
                    "index": index,
                    "heading": f"{collection_name} (item #{index})",
                    "data": item_data,
                }
            )

        # Generate the collection index page
        self._add_file(
            f"data/{dataset_name}/{collection_name}/index.html",
            self._populate_template(
                "templates/collection_index.html",
                collection_name,
                None,
                {"collection_items": items_content},
            ),
        )

    def _add_attachments(self, attachments: list[dict[str, Any]]) -> None:
        """
        Adds top-level attachments to the zip file.

        Args:
            attachments: the attachments to add
        """
        if not attachments or not isinstance(attachments, list):
            return

        # Process attachments and get the links
        attachment_links = self._write_attachment_content(attachments, "attachments")

        # Generate attachments index page using the attachments index template
        self._add_file(
            "attachments/index.html",
            self._populate_template(
                "templates/attachments_index.html",
                "Attachments",
                "Files attached to this privacy request",
                attachment_links,
            ),
        )

    def _get_datasets_from_dsr_data(self) -> dict[str, Any]:
        """
        Returns the datasets from the DSR data.
        """
        # pre-process data to split the dataset:collection keys
        datasets: dict[str, Any] = defaultdict(lambda: defaultdict(list))
        for key, rows in self.dsr_data.items():

            # we handle attachments separately
            if key == "attachments":
                continue

            parts = key.split(":", 1)
            if len(parts) > 1:
                dataset_name, collection_name = parts
            else:
                for row in rows:
                    if "system_name" in row:
                        dataset_name = row["system_name"]
                        collection_name = parts[0]
                        break
                else:
                    dataset_name = "manual"
                    collection_name = parts[0]

            datasets[dataset_name][collection_name].extend(rows)

        return datasets

    def _format_size(self, size_bytes: float) -> str:
        """
        Format size in bytes to human readable format.

        Args:
            size_bytes: Size in bytes

        Returns:
            Formatted string with appropriate unit (B, KB, MB, GB)
        """
        for unit in ["B", "KB", "MB", "GB"]:
            if size_bytes < 1024.0:
                return f"{size_bytes:.1f} {unit}"
            size_bytes /= 1024.0
        return f"{size_bytes:.1f} TB"

    def generate(self) -> BytesIO:
        """
        Processes the request and DSR data to build zip file containing the DSR report.
        Returns the zip file as an in-memory byte array.
        """
        start_time = time_module.time()
        try:
            # all the css for the pages is in main.css
            self._add_file(
                "data/main.css",
                self._populate_template("templates/main.css"),
            )
            self._add_file(
                "data/back.svg",
                Path(os.path.join(DSR_DIRECTORY, "assets/back.svg")).read_text(
                    encoding="utf-8"
                ),
            )

            # pre-process data to split the dataset:collection keys
            datasets: dict[str, Any] = self._get_datasets_from_dsr_data()

            # Sort datasets alphabetically, excluding special cases
            regular_datasets = [
                name for name in sorted(datasets.keys()) if name != "dataset"
            ]  # pylint: disable=invalid-name

            # Add regular datasets in alphabetical order
            for dataset_name in regular_datasets:
                self._add_dataset(dataset_name, datasets[dataset_name])
                self.main_links[dataset_name] = f"data/{dataset_name}/index.html"

            # Add Additional Data if it exists
            if "dataset" in datasets:
                self._add_dataset("dataset", datasets["dataset"])
                self.main_links["Additional Data"] = "data/dataset/index.html"

            # Add Additional Attachments last if it exists
            if "attachments" in self.dsr_data:
                self._add_attachments(self.dsr_data["attachments"])
                self.main_links["Additional Attachments"] = "attachments/index.html"

            # create the main index once all the datasets have been added
            self._add_file(
                "welcome.html",
                self._populate_template(
                    "templates/welcome.html", "DSR Report", None, self.main_links
                ),
            )
        finally:
            # close out zip file in the finally block to always close, even when an exception occurs
            self.out.close()

        # reset the file pointer so the file can be fully read by the caller
        self.baos.seek(0)

        # Calculate time taken and file size
        time_taken = time_module.time() - start_time
        file_size = self._format_size(float(len(self.baos.getvalue())))

        logger.bind(time_to_generate=time_taken, dsr_package_size=file_size).info(
            "DSR report generation complete."
        )
        return self.baos


def _map_privacy_request(privacy_request: PrivacyRequest) -> dict[str, Any]:
    """Creates a map with a subset of values from the privacy request"""
    request_data: dict[str, Any] = {}
    request_data["id"] = privacy_request.id

    action_type: Optional[ActionType] = privacy_request.policy.get_action_type()
    if action_type:
        request_data["type"] = action_type.value

    request_data["identity"] = {
        key: value
        for key, value in privacy_request.get_persisted_identity()
        .labeled_dict(include_default_labels=True)
        .items()
        if value["value"] is not None
    }

    if privacy_request.requested_at:
        request_data["requested_at"] = privacy_request.requested_at.strftime(
            "%m/%d/%Y %H:%M %Z"
        )
    return request_data



from sqlalchemy import Boolean, Column, String
from sqlalchemy.ext.declarative import declared_attr

from fides.api.db.base_class import Base


class CustomConnectorTemplate(Base):
    """A model representing a custom connector template"""

    @declared_attr
    def __tablename__(self) -> str:
        return "custom_connector_template"

    key = Column(String, index=True, unique=True, nullable=False)
    name = Column(String, index=False, unique=False, nullable=False)
    config = Column(String, index=False, unique=False, nullable=False)
    dataset = Column(String, index=False, unique=False, nullable=False)
    icon = Column(String, index=False, unique=False, nullable=True)
    functions = Column(String, index=False, unique=False, nullable=True)
    # indicates that this custom connector template should be replaced
    # if a newer version is available from other sources
    replaceable = Column(
        Boolean, index=False, unique=False, nullable=False, default=False
    )
